from celery import shared_task
from dotenv import load_dotenv
from datetime import datetime
import os
from .utils import get_server_databases, upload_file_to_do_space
from .models import BackupEntry, ContentType
from django.conf import settings

load_dotenv()

@shared_task
def backup_whole_pg_cluster(db_host, db_port, db_name, db_user, db_password, backup_folder, server_id, server_name, content_type_id, locally=False):
    content_type = ContentType.objects.get(id=content_type_id)
    timestamp = datetime.now().strftime("%m_%d_%H-%M-%S")
    cluster_backup_dir = os.path.join(settings.BACKUP_DIR, f"databases/pg_{timestamp}")
    if not os.path.exists(settings.BACKUP_DIR):
        os.mkdir(settings.BACKUP_DIR)
    if not os.path.exists(os.path.join(settings.BACKUP_DIR, 'databases')):
        os.mkdir(os.path.join(settings.BACKUP_DIR, 'databases'))
    if not os.path.exists(cluster_backup_dir):
        os.mkdir(cluster_backup_dir)
    os.environ['PGPASSWORD'] = db_password
    databases = get_server_databases(db_host, db_user, db_password, db_port, db_name)

    backup_entries_to_save = []

    for database in databases:
        print(database, "in process...")
        os.system(f'pg_dump -h {db_host} -U {db_user} -p {db_port} {database} -F c > {os.path.join(backup_folder, f"srv_{server_name}_{database}___{timestamp}.sql")}')
        with open(os.path.join(backup_folder, f"srv_{server_name}_{database}___{timestamp}.sql"), 'rb') as f:
            file_url = upload_file_to_do_space(f, f"srv_{server_name}_{database}___{timestamp}.sql")

            backup_entry = BackupEntry.objects.create(
                content_type=content_type,
                object_id=server_id,
                file=file_url if not locally else f'get_backup/pg_{timestamp}___{database}_.sql'
            )



@shared_task
def backup_database(db_host, db_port, db_name, db_user, db_password, backup_folder, database_id, content_type_id, locally=False):
    content_type = ContentType.objects.get(id=content_type_id)
    timestamp = datetime.now().strftime("%m_%d_%H-%M-%S")
    if not os.path.exists(backup_folder):
        os.mkdir(backup_folder)
    if not os.path.exists(os.path.join(backup_folder, 'databases')):
        os.mkdir(os.path.join(backup_folder, 'databases'))
    os.environ['PGPASSWORD'] = db_password
    os.system(f'pg_dump -h {db_host} -U {db_user} -p {db_port} {db_name} -F c > {os.path.join(backup_folder, f"pg_{timestamp}___{db_name}_.sql")}')

    with open(os.path.join(backup_folder, f'pg_{timestamp}___{db_name}_.sql'), 'rb') as f:
        print("Uploading")
        file_url = upload_file_to_do_space(f, f'pg_{timestamp}___{db_name}_.sql')
        print("Uploaded")
        BackupEntry.objects.create(
            content_type=content_type,
            object_id=database_id,
            file=file_url if not locally else f'get_backup/pg_{timestamp}___{db_name}_.sql'
        )





from django.shortcuts import render
from django.http.response import FileResponse
import os
from django.conf import settings


def get_backup(request, file_name):
    full_file_path = os.path.join(settings.BACKUP_DIR, file_name)
    return FileResponse(open(full_file_path, 'rb'), as_attachment=True, filename=(os.path.basename(file_name)))


# def get_file_dynamic_url(request, file_name):



# Generated by Django 5.0.3 on 2024-09-19 11:56

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Server',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('host', models.CharField(max_length=256)),
                ('default_db_name', models.CharField(default='defaultdb', max_length=128)),
                ('db_user', models.CharField(max_length=256)),
                ('db_password', models.CharField(max_length=256)),
                ('db_port', models.CharField(max_length=256)),
                ('is_accessible', models.BooleanField(default=True)),
                ('last_accessed', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(default='default_server', max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name='BackupEntry',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('file', models.FileField(upload_to='')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
            ],
        ),
        migrations.CreateModel(
            name='Database',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('db_host', models.CharField(max_length=256)),
                ('db_name', models.CharField(max_length=256)),
                ('db_user', models.CharField(max_length=256)),
                ('db_password', models.CharField(max_length=256)),
                ('db_port', models.CharField(max_length=256)),
                ('server', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='databases', to='backups.server')),
            ],
        ),
        migrations.CreateModel(
            name='Table',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256)),
                ('database', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tables', to='backups.database')),
            ],
        ),
        migrations.AddConstraint(
            model_name='database',
            constraint=models.UniqueConstraint(fields=('server', 'db_name'), name='unique_server_db_name_combination'),
        ),
    ]



from django.contrib import admin, messages
from .models import Server, Database, Table, BackupEntry
from .utils import check_connection, get_server_databases, get_database_tables
from .tasks import backup_whole_pg_cluster, backup_database
from django.conf import settings
from django.utils.html import format_html
from django.contrib.contenttypes.models import ContentType

@admin.action(description='check server connection')
def update_connection_status(modeladmin, request, queryset):
    values = queryset.values()[0]
    if check_connection(
        values['host'],
        values['db_user'],
        values['db_password'],
        values['db_port'],
        values['default_db_name'],
    ):
        queryset.update(is_accessible=True)
    else:
        queryset.update(is_accessible=False)

@admin.action(description='get server databases')
def update_server_databases(modeladmin, request, queryset):
    databases_to_create = []
    for server in queryset:
        databases = get_server_databases(
            server.host,
            server.db_user,
            server.db_password,
            server.db_port,
            server.default_db_name
        )
        for database in databases:
            databases_to_create.append(Database(server=server,
                                                db_name=database,
                                                db_host=server.host,
                                                db_user=server.db_user,
                                                db_password=server.db_password,
                                                db_port=server.db_port))
    Database.objects.bulk_create(databases_to_create, ignore_conflicts=True)
    messages.info(request, "Updated databases for selected servers ")

@admin.action(description='Backup all databases')
def backup_server_databases(modeladmin, request, queryset):
    content_type = ContentType.objects.get_for_model(Server)
    for server in queryset:
        args = (
            server.host,
            server.db_port,
            server.default_db_name,
            server.db_user,
            server.db_password,
            str(settings.BACKUP_DIR),
            server.id,
            server.server_name,
            content_type.id
        )
        
        if hasattr(settings, 'CELERY_BROKER_URL'):
            backup_whole_pg_cluster.delay(*args)
        else:
            backup_whole_pg_cluster(*args)

    messages.info(request, "Backup for selected servers has started...")

@admin.action(description='Update all tables')
def update_database_tables(modeladmin, request, queryset):
    tables_to_create = []
    for database in queryset:
        tables = get_database_tables(
            database.db_host,
            database.db_user,
            database.db_password,
            db_port=database.db_port,
            db_name=database.db_name)
        for table in tables:
            tables_to_create.append(Table(database=database, name=table))
    Table.objects.bulk_create(tables_to_create, ignore_conflicts=True)
    messages.info(request, "Updated tables for selected the databases")


@admin.action(description='Backup database')
def backup_databases(modeladmin, request, queryset):
    content_type = ContentType.objects.get_for_model(Database)
    for database in queryset:
        args = (
            database.db_host,
            database.db_port,
            database.db_name,
            database.db_user,
            database.db_password,
            str(settings.BACKUP_DIR),
            database.id,
            content_type.id
        )
        if hasattr(settings, 'CELERY_BROKER_URL'):
            backup_database.delay(
                *args
            )
        else:
            backup_database(*args)
    messages.info(request, "Backup for selected databases has started...")




import psycopg2
import subprocess
import os
from datetime import datetime
import boto3


def check_connection(db_host, db_user, db_password, db_port, db_name):
    try:
        connection = psycopg2.connect(database=db_name, user=db_user, password=db_password, host=db_host, port=db_port, connect_timeout=3)
        cursor = connection.cursor()
        return True
    except Exception as e:
        return False

def get_server_databases(db_host, db_user, db_password, db_port, db_name):
    os.environ['PGPASSWORD'] = db_password
    full_command = f"""psql -t -c "SELECT datname FROM pg_database WHERE datdba = (SELECT oid FROM pg_roles WHERE rolname = \'{db_user}\'); " -h {db_host} -p {db_port} -U {db_user} -d {db_name}"""  # | grep '\\S'"""
    databases = subprocess.check_output(full_command, shell=True, executable='bash').decode().split('\n')
    databases = [i.strip() for i in databases if i != '']
    return databases

def get_database_tables(db_host, db_user, db_password, db_port, db_name):
    os.environ['PGPASSWORD'] = db_password
    full_command = f"""psql -t -c "\\dt" -h {db_host} -p {db_port} -U {db_user} -d {db_name} -P pager=off"""
    entries = subprocess.check_output(full_command, shell=True, executable='bash').decode().split('\n')
    tables = [i.split('|')[1].strip() for i in entries if i != '' and '|' in i]
    return tables

def generate_presigned_url(file_name, expiration=86400):
    s3_client = boto3.client(
        's3',
        aws_access_key_id=os.getenv('DO_SPACE_ACCESS_KEY_ID'),
        aws_secret_access_key=os.getenv('DO_SPACE_SECRET_ACCESS_KEY'),
        region_name=os.getenv('DO_SPACE_REGION'),
        endpoint_url=os.getenv('DO_SPACE_ENDPOINT_URL'),
    )
    try:
        response = s3_client.generate_presigned_url('get_object',
                                                    Params={
                                                        'Bucket': os.getenv('DO_SPACE_BUCKET_NAME'),
                                                        'Key': file_name
                                                    },
                                                    )
    except Exception as e:
        print(f"Error generating pre-signed URL: {e}")
        return None
    return response


def upload_file_to_do_space(file_obj, file_name):

    s3 = boto3.client(
        's3',
        region_name=os.getenv('DO_SPACE_REGION'),
        endpoint_url=os.getenv('DO_SPACE_ENDPOINT_URL'),
        aws_access_key_id=os.getenv('DO_SPACE_ACCESS_KEY_ID'),
        aws_secret_access_key=os.getenv('DO_SPACE_SECRET_ACCESS_KEY')
    )
    s3.upload_fileobj(
        file_obj,
        os.getenv('DO_SPACE_BUCKET_NAME'),
        file_name,
        ExtraArgs={'ACL': 'private'},
        
    )

    file_url = generate_presigned_url(file_name, expiration=3600)

    return file_url



from django.contrib import admin, messages
from .models import Server, Database, Table, BackupEntry
from .actions import *
from django.utils.html import format_html

class ServerInline(admin.TabularInline):
    model = Database

class DatabaseInline(admin.TabularInline):
    model = Table

class ServerAdmin(admin.ModelAdmin):
    list_display = ['host', 'is_accessible']
    ordering = []
    inlines = [ServerInline]
    actions = [update_connection_status, update_server_databases, backup_server_databases]

class DatabaseAdmin(admin.ModelAdmin):
    list_display = ['db_name', 'server_id']
    ordering = []
    inlines = [DatabaseInline]
    actions = [update_database_tables, backup_databases]

class BackupEntryAdmin(admin.ModelAdmin):
    list_display = ['get_content_type_model', 'get_url']

    def get_url(self, obj):
        return format_html(f"""<a href="{obj.file}">{obj.file}</a>""")
    
    def get_content_type_model(self, obj):
        return f"{obj.content_type.model} | {obj.object_id}" if obj.content_type else 'None'
    get_content_type_model.short_description = 'Content Type Model'


class TableAdmin(admin.ModelAdmin):
    list_display = ['name', 'database']
    ordering = []
    actions = [update_database_tables]

admin.site.register(ContentType)
admin.site.register(Table, TableAdmin)
admin.site.register(Server, ServerAdmin)
admin.site.register(Database, DatabaseAdmin)
admin.site.register(BackupEntry, BackupEntryAdmin)




from django.db import models
import psycopg2
from .utils import check_connection
from django.core.exceptions import ValidationError
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType

class Server(models.Model):
    host = models.CharField(max_length=256, null=False, blank=False)
    default_db_name = models.CharField(max_length=128, default="defaultdb", null=False, blank=False)
    db_user = models.CharField(max_length=256, null=False, blank=False)
    db_password = models.CharField(max_length=256, null=False, blank=False)
    db_port = models.CharField(max_length=256, null=False, blank=False)
    is_accessible = models.BooleanField(default=True, null=False, blank=False)
    last_accessed = models.DateTimeField(auto_now_add=True, null=False, blank=False)
    name = models.CharField(max_length=64, default="default_server", null=False, blank=False)

    def save(self, *args, **kwargs):
        if not check_connection(
            db_host=self.host,
            db_user=self.db_user,
            db_password=self.db_password,
            db_port=self.db_port,
            db_name=self.default_db_name
        ):
            raise ValidationError("can't establish connection")

        super(Server, self).save(*args, **kwargs)


class Database(models.Model):
    server = models.ForeignKey(Server, on_delete=models.CASCADE, related_name='databases')
    db_host = models.CharField(max_length=256, null=False, blank=False)
    db_name = models.CharField(max_length=256)
    db_user = models.CharField(max_length=256, null=False, blank=False)
    db_password = models.CharField(max_length=256, null=False, blank=False)
    db_port = models.CharField(max_length=256, null=False, blank=False)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['server', 'db_name'], name='unique_server_db_name_combination'
            )
        ]

    def __str__(self):
        return f"{self.db_name} | {self.server.name}"

class Table(models.Model):
    database = models.ForeignKey(Database, on_delete=models.CASCADE, related_name='tables')
    name = models.CharField(max_length=256)


class BackupEntry(models.Model):
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    file = models.FileField()
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return f"Backup of {self.content_object} created at {self.created_at}"


